!+
! Module: srcmatch_mod
!
! Authors:  Jacco Vink, Frank Jansen - AIP
!          
! Description: cross-identification of XMM sources from EPIC and OM
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module srcmatch_mod

  use dal
  use param
  use ErrorHandling
  use sla
  use srcmatch_def  
  use srcmatch_io

  implicit none 
  
  private copylist, calc_gal_coord, copy_entry, compare_srcs,&
       combine_srcs, crossid_srcs, correlate_om, sort_correlations,&
       init_omcolumns, w_average, dw_average

  public srcmatch

contains 
  !---------------------------------------------------------------------
  subroutine srcmatch
  !---------------------------------------------------------------------
    character(81)       :: emlset, crossidset, omset, summaryset, htmlset
    logical             :: has_ra_corr0,  has_ra_corr
    integer(kind=int32) :: nsrcsets, ncrossid, nsrcs, n_master, nomsrcs
    integer(kind=int32) :: nmax, n_matches
    integer(kind=int32) :: setindex, i
!    real(kind=double)   :: syst_err, max_err, om_dist
    type(Header_type)   :: header, header2
    type(Source_type), dimension(:),pointer :: masterlist,srclist,temp,omlist

    header%syst_err = realParameter("systerr")
    header%max_err = realParameter("maxerr")
    header%om_dist = realParameter("omdist")
    write(errstr,'(A F6.3 A F6.3 A F6.3)') 'sys. error= ',header%syst_err,&
         ' max. error = ',header%max_err,' OM dist. = ',header%om_dist
    call message(AppMsg, Sparse)

    ! epic source list
    nsrcsets=parameterCount("inputlistsets")
    header%nsets =  nsrcsets
    if(nsrcsets > max_nsets) call error('tooManySets','Too many input sets.')

    setindex=0
    emlset = stringParameter("inputlistsets",setindex)
    header%inputsets(1) = emlset
    srclist => read_set(emlset, nsrcs, has_ra_corr0, header)
    write(errstr,'(A A I4 A I4)') emlset,' set nr = ',setindex+1,&
         ' #sources = ',nsrcs
    call message(AppMsg, Sparse)

    !3.7.3 if( nsrcs == 0 ) call error('noData','No data present.')

    ! copy the first dataset to the master list
    if( nsrcsets  == 1 ) then
       nmax = nsrcs
    else
       nmax = 5*nsrcs 
    end if
    if (nmax==0) nmax=500 ! 3.7.3
    ! the master list can contain nmax sources
    masterlist => copylist( srclist, nsrcs, nmax, 1, nsrcsets)
    n_master = nsrcs 

!
!  MERGE SOURCE LIST WITH OTHER SOURCE LISTS
!
    do setindex = 2, nsrcsets
       emlset = stringParameter("inputlistsets",setindex-1)
       header%inputsets(setindex) = emlset
       deallocate(srclist)
       srclist => read_set(emlset, nsrcs, has_ra_corr, header2)
       header%id_inst(setindex) = header2%id_inst(1)
       write(errstr,'(A A I4 A I4)') emlset,'set nr = ',&
            setindex, ' #sources = ',nsrcs
       call message(AppMsg, Sparse)
       n_matches=0
       call compare_srcs(masterlist,n_master, srclist, nsrcs, header%syst_err,&
            header%max_err, n_matches)
       write(errstr,'(A I4 A)') 'Found ',n_matches,' matches'
       call message(AppMsg, Sparse)

       if(  n_matches == 0 ) then
          call warning('noMatchingSources',&
               'No matching sources found between two input sets')
       end if
       if( n_master + (nsrcs-n_matches) > nmax ) then
          call message(AppMsg, Sparse,&
               'Not enough memory available, reallocating...')
          temp => copylist( masterlist, n_master, n_master,0,nsrcsets)
          deallocate( masterlist)
          nmax =  3 * (n_master + (nsrcs-n_matches))
          masterlist => copylist( temp, n_master, nmax,0,nsrcsets)
          deallocate( temp )
       end if
       call message(AppMSG, Sparse, 'Combining the source lists')
       call combine_srcs(masterlist, n_master, srclist, nsrcs, &
            setindex, nsrcsets )
    end do

    ! change the source numbers to the final source numbers
    do i=1,n_master
       masterlist(i)%src_num = i
    end do
!
!  CROSS IDENTIFY WITH STAGE I or II Source lists LIST
!       
    header%ncrossid = 0
    if( booleanParameter("usecrossidsets") ) then
       ncrossid = parameterCount("crossidsets")
       write(errstr,'(A I3)') '#crossid sets = ',ncrossid
       call message(AppMsg, Sparse)

       header%ncrossid =  MIN(ncrossid, max_nsets)
       if(ncrossid > max_nsets) then
          call warning('tooManySets',&
          'Too many input sets, last sets will be ignored')
       end if

       do setindex = 1, ncrossid
          crossidset = stringParameter("crossidsets",setindex-1)
          header%crossidsets(setindex) = crossidset
          deallocate(srclist)
          srclist => read_set(crossidset, nsrcs, has_ra_corr, header2)
          write(errstr,'(A A A I3 A I3)') 'crossidset = ',crossidset,&
               ' set nr = ', setindex,' #sources = ',nsrcs
          call message(AppMsg, Sparse)

          n_matches=0
          call compare_srcs(masterlist,n_master,srclist,nsrcs,header%syst_err,&
               header%max_err, n_matches)
          write(errstr,'(A I3 A)') 'Found ',n_matches,' matches'
          call message(AppMsg, Sparse)

          if(  n_matches == 0 ) then
             call warning('noMatchingSources',&
                  'No matching sources found between two input sets')
          end if
          call crossid_srcs(masterlist, n_master, srclist, nsrcs,setindex)
       end do
    else
       call message(AppMsg, Sparse, 'No cross identfication sets')
       header%ncrossid =  0
    end if

    if( nsrcs > 0 ) deallocate( srclist )

!
!  CROSS IDENTIFY WITH OM LIST
!


    ! OM srcs
    omset = stringParameter("omlistset")
    if(  booleanParameter("useomlistset") ) then
       call init_omcolumns( masterlist, n_master, 'NO_OM_SRCS')
       omlist => read_om_set( omset, nomsrcs)
       write (errstr,'(A A A I4)') 'omset = ',omset, &
            ' # sources = ', nomsrcs
       call message(AppMsg, Sparse)
       call correlate_om(masterlist, n_master, omlist, nomsrcs, header%om_dist)
       call sort_correlations(masterlist, n_master )
       header%omcross = .TRUE.
    else
       header%omcross = .FALSE.
       call message(AppMsg, Sparse,'No OM set will be used')
    end if

    call calc_gal_coord(masterlist, n_master)

!    do i=1,n_master
!       if( i == 1 .OR. i == 4 .OR. i == 70 ) then
!       print "(A F10.6 A F10.6 A F10.6 A F10.6)",'ra = ' , masterlist(i)%ra, &
!               ' dec = ',  masterlist(i)%dec,' lII = ', masterlist(i)%lii, &
!               ' bII = ', masterlist(i)%bii
!       end if
!    end do


!
! NB write only the RA/DEC_CORR columns when only one dataset is involved
! for reasons of complexity
!
    summaryset = stringParameter("outputlistset")
    if( nsrcsets == 1) then 
       call write_summary_set(summaryset, masterlist, n_master, has_ra_corr0,&
            header,nsrcsets)
    else
       call write_summary_set(summaryset, masterlist, n_master, .FALSE.,&
            header,nsrcsets)
    end if
    htmlset = stringParameter("htmloutput")
    call  write_html_list(htmlset,masterlist, n_master,header, nsrcsets)

    if( n_master > 0 ) deallocate( masterlist)

  end subroutine srcmatch


  !---------------------------------------------------------------------
  subroutine compare_srcs(list1, n1, list2, n2, systerr, maxerror, nmatches )
  !---------------------------------------------------------------------
    !
    ! The two lists (list1/2) are compared, and the nearest counterparts
    ! are linked by references. If a new counterpart is encountered, which
    ! is nearer than an earlier found counterpart, the old link is destroyed
    ! and a new link made.
    !
    type(Source_type), dimension(:), intent(in out) :: list1, list2
    integer(kind=int32),intent(in) :: n1, n2
    integer(kind=int32),intent(out) :: nmatches
    real(kind=double), intent(in) :: systerr, maxerror

    integer(kind=int32) :: i, j, ref, ref1, ref2
    real(kind=double) :: angdist, deltaRA, cosdist, radec_err

    !3.7.3 if( n1 <= 0 .OR. n2 <= 0 ) then 
    !        call error('emptySet','Input source lists are empty.')
    !3.7.3 end if

    do i=1,n1
       list1(i)%cross_ref = null_entry
    end do
    do j=1,n2
       list2(j)%cross_ref = null_entry
    end do

    do i=1,n1
       do j=1,n2

          deltaRA = (list2(j)%ra - list1(i)%ra) * deg2rad
          cosdist = dsin(list1(i)%dec*deg2rad) * dsin(list2(j)%dec*deg2rad) + &
               dcos(list1(i)%dec*deg2rad)*dcos(list2(j)%dec*deg2rad) * &
               dcos(deltaRA)
          ! This is strange but necessary!!!
          if( cosdist > 0.9 ) cosdist = min(1.0d0, cosdist) 
          if( cosdist < -0.9 ) cosdist = max(-1.0d0, cosdist) 

          angdist = dacos(cosdist)*rad2arcsec
          radec_err =  sqrt( (list1(i)%radec_err)**2+(list2(j)%radec_err)**2 +&
                 systerr**2)
          angdist = angdist/radec_err        ! so it is a relative angular dist
          ref1 = list1(i)%cross_ref
          ref2 = list2(j)%cross_ref
          if( angdist < maxerror .AND. ref1 /= j .AND. ref2 /= i) then
             if( ref1 == 0 .AND. ref2 == 0) then
                list1(i)%cross_ref = j
                list2(j)%cross_ref = i
                list1(i)%dist = angdist
                list2(j)%dist = angdist
             else 
                if(ref1 /= 0 .AND. ref2 == 0 .AND. list1(i)%dist > angdist)then
                   list2(ref1)%cross_ref = 0 ! old association
                   list1(i)%cross_ref = j
                   list2(j)%cross_ref = i
                   list1(i)%dist = angdist
                   list2(j)%dist = angdist
                else
                   if( ref1 == 0 .AND. ref2 /= 0 .AND. &
                        list2(j)%dist > angdist )then
                      ref = list1(ref2)%cross_ref
                      list1(ref2)%cross_ref = 0 ! old association
                      list1(i)%cross_ref = j
                      list2(j)%cross_ref = i
                      list1(i)%dist = angdist
                      list2(j)%dist = angdist
                   else
                      if( ref1 /= 0 .AND. ref2 /= 0 .AND. &
                           list1(i)%dist > angdist .AND. &
                           list2(j)%dist > angdist )then
                         list1(ref2)%cross_ref = 0
                         list2(ref1)%cross_ref = 0 ! remove old associations
                         list1(i)%cross_ref = j
                         list2(j)%cross_ref = i
                         list1(i)%dist = angdist
                         list2(j)%dist = angdist
                      end if
                   end if
                end if
             end if
          end if
!          if( i == j .AND. list1(i)%cross_ref == 0 ) then
!             print "(I4 I4 A F8.2 A E12.4 A I4 A I4 A E12.4)",i,j,' d=  ',angdist,&
!                  ' cosdist -1 = ',cosdist-1, '  ',&
!                  list1(i)%cross_ref,'  ',list2(j)%cross_ref, &
!                  ' diff ',list1(i)%ra - list2(j)%ra1
!
!          end if
       end do
    end do

    nmatches = 0
    do i=1,n1
       if( list1(i)%cross_ref /= 0 ) nmatches = nmatches + 1
    end do

!    nmatches = 0
!    do i=1,n2
!       if( list2(i)%cross_ref /= 0 ) nmatches = nmatches + 1
!    end do


  end subroutine compare_srcs


  !---------------------------------------------------------------------
  subroutine combine_srcs(list1, n1, list2, n2, setnr, nsets)
  !---------------------------------------------------------------------
    type(Source_type), dimension(:), intent(in out) :: list1
    type(Source_type), dimension(:), intent(in) :: list2
    integer(kind=int32),intent(in out) :: n1
    integer(kind=int32),intent(in) :: n2,setnr,nsets

    integer(kind=int32) :: i, j, ref, oldn1
    real(kind=double) :: newradec_err


    oldn1 = n1
    do i=1, n2
       ref = list2(i)%cross_ref

       if( ref /= null_entry ) then
          !
          ! We should attach the new information to the already existing srcnr
          !
          ! weighted average of some quantities
          call dw_average( list1(ref)%ra, real(list1(ref)%radec_err,2), &
               list2(i)%ra,real(list2(i)%radec_err,2),&
               list1(ref)%ra, newradec_err)   
          call dw_average( list1(ref)%dec, real(list1(ref)%radec_err,2), &
               list2(i)%dec, real(list2(i)%radec_err,2),&
               list1(ref)%dec,newradec_err)
          list1(ref)%radec_err = real(newradec_err,1)

          call w_average( list1(ref)%pn_tot, list1(ref)%pn_tot_err, &
                          list2(i)%pn_tot,list2(i)%pn_tot_err, &
                          list1(ref)%pn_tot, list1(ref)%pn_tot_err)
          call w_average( list1(ref)%m1_tot, list1(ref)%m1_tot_err, &
                          list2(i)%m1_tot,list2(i)%m1_tot_err, &
                          list1(ref)%m1_tot, list1(ref)%m1_tot_err)
          call w_average( list1(ref)%m2_tot, list1(ref)%m2_tot_err, &
                          list2(i)%m2_tot,list2(i)%m2_tot_err, &
                          list1(ref)%m2_tot, list1(ref)%m2_tot_err)
          call w_average( list1(ref)%ep_tot, list1(ref)%ep_tot_err, &
                          list2(i)%ep_tot,list2(i)%ep_tot_err, &
                          list1(ref)%ep_tot, list1(ref)%ep_tot_err)

          call w_average( list1(ref)%pn_xid, list1(ref)%pn_xid_err, &
                          list2(i)%pn_xid,list2(i)%pn_xid_err, &
                          list1(ref)%pn_xid, list1(ref)%pn_xid_err)
          call w_average( list1(ref)%m1_xid, list1(ref)%m1_xid_err, &
                          list2(i)%m1_xid,list2(i)%m1_xid_err, &
                          list1(ref)%m1_xid, list1(ref)%m1_xid_err)
          call w_average( list1(ref)%m2_xid, list1(ref)%m2_xid_err, &
                          list2(i)%m2_xid,list2(i)%m2_xid_err, &
                          list1(ref)%m2_xid, list1(ref)%m2_xid_err)
          call w_average( list1(ref)%ep_xid, list1(ref)%ep_xid_err, &
                          list2(i)%ep_xid,list2(i)%ep_xid_err, &
                          list1(ref)%ep_xid, list1(ref)%ep_xid_err)

          ! Harness ratios
          call w_average( list1(ref)%pn_hr1, list1(ref)%pn_hr1_err, &
                          list2(i)%pn_hr1,list2(i)%pn_hr1_err, &
                          list1(ref)%pn_hr1, list1(ref)%pn_hr1_err)
          call w_average( list1(ref)%pn_hr2, list1(ref)%pn_hr2_err, &
                          list2(i)%pn_hr2,list2(i)%pn_hr2_err, &
                          list1(ref)%pn_hr2, list1(ref)%pn_hr2_err)
          call w_average( list1(ref)%pn_hr3, list1(ref)%pn_hr3_err, &
                          list2(i)%pn_hr3,list2(i)%pn_hr3_err, &
                          list1(ref)%pn_hr3, list1(ref)%pn_hr3_err)

          call w_average( list1(ref)%m1_hr1, list1(ref)%m1_hr1_err, &
                          list2(i)%m1_hr1,list2(i)%m1_hr1_err, &
                          list1(ref)%m1_hr1, list1(ref)%m1_hr1_err)
          call w_average( list1(ref)%m1_hr2, list1(ref)%m1_hr2_err, &
                          list2(i)%m1_hr2,list2(i)%m1_hr2_err, &
                          list1(ref)%m1_hr2, list1(ref)%m1_hr2_err)
          call w_average( list1(ref)%m1_hr3, list1(ref)%m1_hr3_err, &
                          list2(i)%m1_hr3,    list2(i)%m1_hr3_err, &
                          list1(ref)%m1_hr3, list1(ref)%m1_hr3_err)

          call w_average( list1(ref)%m2_hr1, list1(ref)%m2_hr1_err, &
                          list2(i)%m2_hr1,list2(i)%m2_hr1_err, &
                          list1(ref)%m2_hr1, list1(ref)%m2_hr1_err)
          call w_average( list1(ref)%m2_hr2, list1(ref)%m2_hr2_err, &
                          list2(i)%m2_hr2,list2(i)%m2_hr2_err, &
                          list1(ref)%m2_hr2, list1(ref)%m2_hr2_err)
          call w_average( list1(ref)%m2_hr3, list1(ref)%m2_hr3_err, &
                          list2(i)%m2_hr3,list2(i)%m2_hr3_err, &
                          list1(ref)%m2_hr3, list1(ref)%m2_hr3_err)

!
! NB should those be copied, added or avaraged!!!!
!
          call w_average( list1(ref)%pn_extent, list1(ref)%pn_ext_err, &
                          list2(i)%pn_extent,list2(i)%pn_ext_err, &
                          list1(ref)%pn_extent, list1(ref)%pn_ext_err)
          call w_average( list1(ref)%m1_extent, list1(ref)%m1_ext_err, &
                          list2(i)%m1_extent,list2(i)%m1_ext_err, &
                          list1(ref)%m1_extent, list1(ref)%m1_ext_err)
          call w_average( list1(ref)%m2_extent, list1(ref)%m2_ext_err, &
                          list2(i)%m2_extent,list2(i)%m2_ext_err, &
                          list1(ref)%m2_extent, list1(ref)%m2_ext_err)
          call w_average( list1(ref)%ep_extent, list1(ref)%ep_ext_err, &
                          list2(i)%ep_extent,list2(i)%ep_ext_err, &
                          list1(ref)%ep_extent, list1(ref)%ep_ext_err)
!

          list1(ref)%ep_ml_id_src(setnr) = list2(i)%ep_ml_id_src(1)
          list1(ref)%pn_ml_id_src(setnr) = list2(i)%pn_ml_id_src(1)
          list1(ref)%m1_ml_id_src(setnr) = list2(i)%m1_ml_id_src(1)
          list1(ref)%m2_ml_id_src(setnr) = list2(i)%m2_ml_id_src(1)
          list1(ref)%ep_box_id_src(setnr) = list2(i)%ep_box_id_src(1)
          list1(ref)%pn_box_id_src(setnr) = list2(i)%pn_box_id_src(1)
          list1(ref)%m1_box_id_src(setnr) = list2(i)%m1_box_id_src(1)
          list1(ref)%m2_box_id_src(setnr) = list2(i)%m2_box_id_src(1)
          list1(ref)%pn_det_ml(setnr) = list2(i)%pn_det_ml(1)
          list1(ref)%m1_det_ml(setnr) = list2(i)%m1_det_ml(1)
          list1(ref)%m2_det_ml(setnr) = list2(i)%m2_det_ml(1)
          list1(ref)%ep_det_ml(setnr) = list2(i)%ep_det_ml(1)
          list1(ref)%pn_ext_ml(setnr) = list2(i)%pn_ext_ml(1)
          list1(ref)%m1_ext_ml(setnr) = list2(i)%m1_ext_ml(1)
          list1(ref)%m2_ext_ml(setnr) = list2(i)%m2_ext_ml(1)
          list1(ref)%ep_ext_ml(setnr) = list2(i)%ep_ext_ml(1)
       else
          !
          ! just put the new entry at the end of the list
          n1 = n1 + 1
          do j=1,setnr-1                    
             list1(n1)%ep_ml_id_src(j) = null_entry
             list1(n1)%pn_ml_id_src(j) = null_entry
             list1(n1)%m1_ml_id_src(j) = null_entry
             list1(n1)%m2_ml_id_src(j) = null_entry
             list1(n1)%ep_box_id_src(j) = null_entry
             list1(n1)%pn_box_id_src(j) = null_entry
             list1(n1)%m1_box_id_src(j) = null_entry
             list1(n1)%m2_box_id_src(j) = null_entry
          end do
          list1(n1)%pn_det_ml(setnr) = 0.0
          list1(n1)%m1_det_ml(setnr) = 0.0
          list1(n1)%m2_det_ml(setnr) = 0.0
          list1(n1)%ep_det_ml(setnr) = 0.0
          list1(n1)%pn_ext_ml(setnr) = 0.0
          list1(n1)%m1_ext_ml(setnr) = 0.0
          list1(n1)%m2_ext_ml(setnr) = 0.0
          list1(n1)%ep_ext_ml(setnr) = 0.0
          call copy_entry( list1(n1), list2(i), setnr, nsets )
       end if
    end do
    !print "(A I3 A I3)", ' old n1 = ',oldn1, ' new n1 = ',n1
  end subroutine combine_srcs


  !---------------------------------------------------------------------
  subroutine crossid_srcs(list1, n1, list2, n2, setnr)
  !---------------------------------------------------------------------
    type(Source_type), dimension(:), intent(in out) :: list1
    type(Source_type), dimension(:), intent(in) :: list2
    integer(kind=int32),intent(in out) :: n1
    integer(kind=int32),intent(in) :: n2,setnr

    integer(kind=int32) :: i, ref

    do i=1, n1
       ref = list1(i)%cross_ref
       if( ref /= 0 ) then
          list1(i)%pn_ml_id_src(setnr) = list2(ref)%pn_ml_id_src(1)
          list1(i)%m1_ml_id_src(setnr) = list2(ref)%m1_ml_id_src(1)
          list1(i)%m2_ml_id_src(setnr) = list2(ref)%m2_ml_id_src(1)
          list1(i)%pn_box_id_src(setnr) = list2(ref)%pn_box_id_src(1)
          list1(i)%m1_box_id_src(setnr) = list2(ref)%m1_box_id_src(1)
          list1(i)%m2_box_id_src(setnr) = list2(ref)%m2_box_id_src(1)
       else
          list1(i)%pn_ml_id_src(setnr) =  null_entry
          list1(i)%m1_ml_id_src(setnr) =  null_entry
          list1(i)%m2_ml_id_src(setnr) =  null_entry
          list1(i)%pn_box_id_src(setnr) = null_entry
          list1(i)%m1_box_id_src(setnr) = null_entry
          list1(i)%m2_box_id_src(setnr) = null_entry
       end if
    end do

  end subroutine crossid_srcs



  !---------------------------------------------------------------------
  function copylist( inputlist, nsrcs, nmaxsrcs, setnr, nsets) &
       result(output_listptr)
  !---------------------------------------------------------------------       
    type(Source_type), dimension(:), pointer :: output_listptr
    type(Source_type), dimension(:), intent(in) :: inputlist
    integer(kind=int32) , intent (in) :: nsrcs, nmaxsrcs, setnr, nsets

    integer(kind=int32) :: i

    !3.7.3 if( nsrcs <= 0 ) then
    !        call error('noData','No data present.')
    !3.7.3 end if

!The following condition should not happen, so I don't test for it.
!    if( nmaxsrcs < nsrcs ) then
!       call error('outOfMemory','Not enough memory is allocated.')
!    end if


    allocate(output_listptr(nmaxsrcs))
    call init_list(output_listptr,nmaxsrcs)

    do i=1, nsrcs
       call copy_entry(  output_listptr(i), inputlist(i),setnr, nsets )
    end do

    if( nmaxsrcs > nsrcs ) then
       do i=nsrcs+1, nmaxsrcs
          output_listptr(i)%ra      = 0.0
          output_listptr(i)%dec     = 0.0
          output_listptr(i)%radec_err = 0.0
       end do
    end if

  end function copylist

  !---------------------------------------------------------------------
  subroutine copy_entry( entry1, entry2, setnr, nsets)
  !---------------------------------------------------------------------
    ! copies entry2 to entry1
    type(Source_type),intent(in out) :: entry1
    type(Source_type),intent(in)   :: entry2
    integer(kind=int32),intent(in) :: setnr,nsets

    integer(kind=int32) :: i

    !
    ! setnr!= 0: copy all but for the src numbers use index setnr
    ! setnr ==0 : (literal) copy all with a loop over the index 
    !
    entry1%src_num  =  entry2%src_num
    if( setnr /= 0 ) then
       entry1%ep_ml_id_src(setnr) =  entry2%ep_ml_id_src(1)
       entry1%pn_ml_id_src(setnr) =  entry2%pn_ml_id_src(1)
       entry1%m1_ml_id_src(setnr) =  entry2%m1_ml_id_src(1)
       entry1%m2_ml_id_src(setnr) =  entry2%m2_ml_id_src(1)
       entry1%ep_box_id_src(setnr) =  entry2%ep_box_id_src(1)
       entry1%pn_box_id_src(setnr) =  entry2%pn_box_id_src(1)
       entry1%m1_box_id_src(setnr) =  entry2%m1_box_id_src(1)
       entry1%m2_box_id_src(setnr) =  entry2%m2_box_id_src(1)
       ! max. likelihood values
       entry1%pn_det_ml(setnr)   =  entry2%pn_det_ml(1)
       entry1%m1_det_ml(setnr)   =  entry2%m1_det_ml(1)
       entry1%m2_det_ml(setnr)   =  entry2%m2_det_ml(1)
       entry1%ep_det_ml(setnr)   =  entry2%ep_det_ml(1)
       entry1%pn_ext_ml(setnr)   =  entry2%pn_ext_ml(1)
       entry1%m1_ext_ml(setnr)   =  entry2%m1_ext_ml(1)
       entry1%m2_ext_ml(setnr)   =  entry2%m2_ext_ml(1)
       entry1%ep_ext_ml(setnr)   =  entry2%ep_ext_ml(1)
    else
       do i=1,nsets
          entry1%ep_ml_id_src(i) =  entry2%ep_ml_id_src(i)
          entry1%pn_ml_id_src(i) =  entry2%pn_ml_id_src(i)
          entry1%m1_ml_id_src(i) =  entry2%m1_ml_id_src(i)
          entry1%m2_ml_id_src(i) =  entry2%m2_ml_id_src(i)
          entry1%ep_box_id_src(i) =  entry2%ep_box_id_src(i)
          entry1%pn_box_id_src(i) =  entry2%pn_box_id_src(i)
          entry1%m1_box_id_src(i) =  entry2%m1_box_id_src(i)
          ! max. likelihood values
          entry1%pn_det_ml(i)   =  entry2%pn_det_ml(i)
          entry1%m1_det_ml(i)   =  entry2%m1_det_ml(i)
          entry1%m2_det_ml(i)   =  entry2%m2_det_ml(i)
          entry1%ep_det_ml(i)   =  entry2%ep_det_ml(i)
          entry1%pn_ext_ml(i)   =  entry2%pn_ext_ml(i)
          entry1%m1_ext_ml(i)   =  entry2%m1_ext_ml(i)
          entry1%m2_ext_ml(i)   =  entry2%m2_ext_ml(i)
          entry1%ep_ext_ml(i)   =  entry2%ep_ext_ml(i)
       end do
    end if

    entry1%ra  =  entry2%ra
    entry1%dec =  entry2%dec
    entry1%ra_corr  =  entry2%ra_corr
    entry1%dec_corr =  entry2%dec_corr
    entry1%radec_err =  entry2%radec_err
    entry1%lii =  entry2%lii
    entry1%bii =  entry2%bii
    entry1%lii_corr =  entry2%lii_corr
    entry1%bii_corr =  entry2%bii_corr

    ! countrates + errors:
    entry1%pn_tot      =  entry2%pn_tot
    entry1%pn_tot_err  =  entry2%pn_tot_err
    entry1%m1_tot      =  entry2%m1_tot
    entry1%m1_tot_err  =  entry2%m1_tot_err
    entry1%m2_tot      =  entry2%m2_tot
    entry1%m2_tot_err  =  entry2%m2_tot_err
    entry1%ep_tot      =  entry2%ep_tot
    entry1%ep_tot_err  =  entry2%ep_tot_err
    entry1%pn_xid      =  entry2%pn_xid
    entry1%pn_xid_err  =  entry2%pn_xid_err
    entry1%m1_xid      =  entry2%m1_xid
    entry1%m1_xid_err  =  entry2%m1_xid_err
    entry1%m2_xid      =  entry2%m2_xid
    entry1%m2_xid_err  =  entry2%m2_xid_err
    entry1%ep_xid      =  entry2%ep_xid
    entry1%ep_xid_err  =  entry2%ep_xid_err


    ! extent parameters
    entry1%pn_extent   =  entry2%pn_extent
    entry1%pn_ext_err  =  entry2%pn_ext_err
    entry1%m1_extent   =  entry2%m1_extent
    entry1%m1_ext_err  =  entry2%m1_ext_err
    entry1%m2_extent   =  entry2%m2_extent
    entry1%m2_ext_err  =  entry2%m2_ext_err
    entry1%ep_extent   =  entry2%ep_extent
    entry1%ep_ext_err  =  entry2%ep_ext_err

    ! hardness ratios
    entry1%pn_hr1      =  entry2%pn_hr1
    entry1%pn_hr1_err  =  entry2%pn_hr1_err
    entry1%pn_hr2      =  entry2%pn_hr2
    entry1%pn_hr2_err  =  entry2%pn_hr2_err
    entry1%pn_hr3      =  entry2%pn_hr3
    entry1%pn_hr3_err  =  entry2%pn_hr3_err
    entry1%m1_hr1      =  entry2%m1_hr1
    entry1%m1_hr1_err  =  entry2%m1_hr1_err
    entry1%m1_hr2      =  entry2%m1_hr2
    entry1%m1_hr2_err  =  entry2%m1_hr2_err
    entry1%m1_hr3      =  entry2%m1_hr3
    entry1%m1_hr3_err  =  entry2%m1_hr3_err
    entry1%m2_hr1      =  entry2%m2_hr1
    entry1%m2_hr1_err  =  entry2%m2_hr1_err
    entry1%m2_hr2      =  entry2%m2_hr2
    entry1%m2_hr2_err  =  entry2%m2_hr2_err
    entry1%m2_hr3      =  entry2%m2_hr3
    entry1%m2_hr3_err  =  entry2%m2_hr3_err
    entry1%ep_hr1      =  entry2%ep_hr1
    entry1%ep_hr1_err  =  entry2%ep_hr1_err
    entry1%ep_hr2      =  entry2%ep_hr2
    entry1%ep_hr2_err  =  entry2%ep_hr2_err
    entry1%ep_hr3      =  entry2%ep_hr3
    entry1%ep_hr3_err  =  entry2%ep_hr3_err

    ! other
    do i=1,5
       entry1%om_id(i) =  entry2%om_id(i)
       entry1%om_id_dist(i) =  entry2%om_id_dist(i)
    end do
    entry1%n_opt_co =  entry2%n_opt_co
    entry1%omflag   =  entry2%omflag

    entry1%spectra  =  entry2%spectra
    entry1%tseries  =  entry2%tseries
    entry1%val_flag =  entry2%val_flag
    entry1%ver_flag =  entry2%ver_flag
    entry1%ver_comm =  entry2%ver_comm

    ! etc

  end subroutine copy_entry


  !---------------------------------------------------------------------
  subroutine correlate_om( srclist, nsrcs, omlist, nom, omdist)
  !---------------------------------------------------------------------   
    type(Source_type), dimension(:), intent(in out) :: srclist, omlist
    integer(kind=int32) , intent (in) :: nsrcs, nom
    real(kind=double), intent(in) :: omdist

    integer(kind=int32) :: i, j, k, kmin
    real(kind=double)   :: deltaRA, cosdist, angdist, maxdist

    do i=1,nsrcs
       do j=1,nom

          deltaRA = (omlist(j)%ra - srclist(i)%ra) * deg2rad
          cosdist = dsin(srclist(i)%dec*deg2rad)*dsin(omlist(j)%dec*deg2rad) +&
               dcos(srclist(i)%dec*deg2rad)*dcos(omlist(j)%dec*deg2rad) * &
               dcos(deltaRA)
          ! This is strange but necessary!!!
          if( cosdist > 0.8 ) cosdist = min(1.0d0, cosdist) 
          if( cosdist < -0.8 ) cosdist = max(-1.0d0, cosdist) 
          angdist = dacos(cosdist)*rad2arcsec

          if( angdist < omdist ) then
             srclist(i)%n_opt_co =  srclist(i)%n_opt_co + 1
             if( srclist(i)%n_opt_co <= 5 ) then
                 srclist(i)%om_id(srclist(i)%n_opt_co) = omlist(j)%src_num
                 srclist(i)%om_id_dist(srclist(i)%n_opt_co) = angdist
                 if( srclist(i)%n_opt_co == 1 )  srclist(i)%omflag='OM_SRC_FND'
             else
                maxdist= srclist(i)%om_id_dist(1)
                kmin=1
                do k=2,5
                   if( srclist(i)%om_id_dist(k) > maxdist ) then
                      maxdist =  srclist(i)%om_id_dist(k) 
                      kmin=k
                   end if
                end do
                if(  maxdist > angdist ) then
                   srclist(i)%om_id(kmin) = omlist(j)%src_num
                   srclist(i)%om_id_dist(kmin) = angdist
                end if
             end if
          end if
       end do
    end do


  end subroutine correlate_om

  !---------------------------------------------------------------------
  subroutine sort_correlations(srclist, nsrcs)
  !--------------------------------------------------------------------- 
    type(Source_type), dimension(:), intent(in out) :: srclist
    integer(kind=int32) , intent (in) :: nsrcs
    integer(kind=int32) :: i, j,k, kmin, id, nopt
    real(kind=single) :: dist

    do i=1,nsrcs
       if(srclist(i)%n_opt_co > 1 ) then
          nopt = MIN(srclist(i)%n_opt_co,5)
          do k=1,nopt-1
             kmin=k
             do j=kmin+1,nopt
                if(srclist(i)%om_id_dist(j) < srclist(i)%om_id_dist(kmin)) then
                   kmin=j
                end if
             end do
!             print "(I4)", kmin
             if( kmin /= k ) then
                !swap
                id = srclist(i)%om_id(k)
                dist = srclist(i)%om_id_dist(k)
                srclist(i)%om_id_dist(k) =  srclist(i)%om_id_dist(kmin)
                srclist(i)%om_id(k) =  srclist(i)%om_id(kmin)
                srclist(i)%om_id_dist(kmin) = dist
                srclist(i)%om_id(kmin) = id
             end if
          end do
       end if
    end do

  end subroutine sort_correlations

  !---------------------------------------------------------------------
  subroutine calc_gal_coord( srclist, nsrcs)
  !---------------------------------------------------------------------       
    type(Source_type), dimension(:), intent(in out) :: srclist
    integer(kind=int32) , intent (in) :: nsrcs
    integer(kind=int32):: i
    real(kind=double) :: alpha, delta, b,l
    real(kind=double) :: cosdecgp, sindecgp

!
! use cosine formual to calculate b (NB cos(90-delta)=sin(delta)
! and use cosine + sine formual to find l

    cosdecgp =  dcos( dec_gal_pole * deg2rad)
    sindecgp =  dsin( dec_gal_pole * deg2rad)

    do i=1,nsrcs
       alpha = srclist(i)%ra
       delta = srclist(i)%dec
       call sla_EQGAL (srclist(i)%ra*deg2rad , srclist(i)%dec*deg2rad,&
            l, b)
       srclist(i)%lii = l*rad2deg
       srclist(i)%bii = b*rad2deg
       call sla_EQGAL(srclist(i)%ra_corr*deg2rad, srclist(i)%dec_corr*deg2rad,&
            l, b)
       srclist(i)%lii_corr = l*rad2deg
       srclist(i)%bii_corr = b*rad2deg
    end do

  end subroutine calc_gal_coord


  !---------------------------------------------------------------------
  subroutine init_omcolumns( srclist, nsrcs, flag)
  !---------------------------------------------------------------------    
    type(Source_type), dimension(:), intent(in out) :: srclist
    integer(kind=int32),intent(in) :: nsrcs
    character(10),intent(in)       :: flag
    integer(kind=int32) :: i,j

    do i=1,nsrcs
       srclist(i)%omflag = flag
       srclist(i)%n_opt_co = 0
       do j=1,5
          srclist(i)%om_id(j)  = 0
          srclist(i)%om_id_dist(j)= 0.0
       end do
    end do

  end subroutine init_omcolumns

  !---------------------------------------------------------------------
  subroutine w_average( value1, error1, value2, error2, av_val, av_err)
  !---------------------------------------------------------------------    
    real(kind=single), intent(in) :: value1, error1, value2, error2
    real(kind=single), intent(out) :: av_val, av_err
    real(kind=single) :: w1, w2, totw, newval, newerr ! weights

    if(.not. error1 > 0.0 ) then
       av_val = value2
       av_err = error2
    else
       if(.not. error2 > 0.0 ) then
          av_val = value1
          av_err = error1
       else
          ! error1 /= 0 && error2 /= 0: weighted average
          w1 =  1.0/error1**2
          w2 =  1.0/error2**2
          totw = w1 + w2
          newval = (w1 * value1 + w2 * value2)/totw
          newerr = 1.0/sqrt( totw )
          av_val = newval
          av_err = newerr
       end if
    end if

  end subroutine w_average

  ! SAME as w_average , but now for double precission
  ! 
  !---------------------------------------------------------------------
  subroutine dw_average( value1, error1, value2, error2, av_val, av_err)
  !---------------------------------------------------------------------    
    real(kind=double), intent(in) :: value1, error1, value2, error2
    real(kind=double), intent(out) :: av_val, av_err
    real(kind=double) :: w1, w2, totw, newval, newerr ! weights

    if(.not. error1 > 0.0 ) then
       av_val = value2
       av_err = error2
    else
       if(.not. error2 > 0.0 ) then
          av_val = value1
          av_err = error1
       else
          ! error1 /= 0 && error2 /= 0: weighted average
          w1 =  1.0/error1**2
          w2 =  1.0/error2**2
          totw = w1 + w2
          newval = (w1 * value1 + w2 * value2)/totw
          newerr = 1.0/sqrt( totw )
          av_val = newval
          av_err = newerr
       end if
    end if

  end subroutine dw_average

end module srcmatch_mod
